//LUT_EVAL - 1D value binning given a sorted list of thresholds (supports linear interpolation and "circular" bins)
//
//   Compile with MAKEMEX.
//
//   Author: Uwe Schmidt, TU Darmstadt (uwe.schmidt@gris.tu-darmstadt.de)
//
//   This file is part of the implementation as described in the CVPR 2012 paper:
//   Uwe Schmidt and Stefan Roth. Learning Rotation-Aware Features: From Invariant Priors to Equivariant Descriptors.
//   Please see the file LICENSE.txt for the license governing this code.

#include <math.h>
#include "mex.h"

#ifdef _MSC_VER
#include <float.h>
#define isfinite _finite
#endif


static inline int sign(double x) { return (x < 0 ? -1 : (x > 0 ? 1 : 0)); }

int binary_search(const double v, const double *thr, int e) {
  int m, s = 0;
  while (s < e) {
    m = s + (e-s)/2;
    if (v <= *(thr+m)) e = m; else s = m+1;
  }
  return s;
}


void binning(double *B, double *Bidx, const double *I, const int nelements, const double *thr, const int nbins, const int do_interp, const int wrap) {

  double * bin_widths  = (double*) mxMalloc(nbins * sizeof(double));
  double * bin_centers = (double*) mxMalloc(nbins * sizeof(double));

  for (int i = 0; i < nbins; i++) {
    bin_widths[i] = thr[i+1]-thr[i];
    bin_centers[i] = thr[i] + bin_widths[i]/2;
  }
  
  for (int i = 0; i < nelements; i++) {
    const double v = *(I+i);
    // element out of bounds
    if (v < thr[0] || v > thr[nbins]) *(Bidx+i) = -1;
    else {
      const int bin = binary_search(v, thr+1, nbins-1);
      *(Bidx+i) = (double) bin + 1; // convert back to matlab index counting
      // no interpolation
      if (!do_interp) {
        *(B + nelements*bin + i) = 1;
      // do interpolation...
      } else {
        const double pos_from_bin_center = v - bin_centers[bin];
        const int neighbor_offset = sign(pos_from_bin_center);
        // ...with circular bins
        if (neighbor_offset != 0 && wrap) {
          const double rel_pos_from_bin_center = pos_from_bin_center / bin_widths[0];
          const double distance_from_bin_center = fabs(rel_pos_from_bin_center);
          const int bin_neighbor = (nbins + bin+neighbor_offset) % nbins;
          *(B + nelements*bin          + i) = 1-distance_from_bin_center;
          *(B + nelements*bin_neighbor + i) =   distance_from_bin_center;
        // ...without circular bins
        } else if (neighbor_offset != 0         &&
                   isfinite(bin_centers[bin])   &&
                   bin+neighbor_offset >= 0     &&
                   bin+neighbor_offset < nbins  &&
                   isfinite(bin_centers[bin+neighbor_offset])) {
          const double dist_neighboring_bin_centers = fabs(bin_centers[bin+neighbor_offset] - bin_centers[bin]);
          const double rel_pos_from_bin_center = pos_from_bin_center / dist_neighboring_bin_centers;
          const double distance_from_bin_center = fabs(rel_pos_from_bin_center);
          *(B + nelements*bin                   + i) = 1-distance_from_bin_center;
          *(B + nelements*(bin+neighbor_offset) + i) =   distance_from_bin_center;
        // ...not possible
        } else {
          *(B + nelements*bin + i) = 1;
        }
      }
    }
  }

  mxFree(bin_widths);
  mxFree(bin_centers);
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  
  if( nrhs<2 || nrhs>4 ) mexErrMsgTxt("2-4 inputs expected.");
  if( nlhs>2 ) mexErrMsgTxt("1-2 outputs expected.");
  if( mxGetClassID(prhs[0])!=mxDOUBLE_CLASS || mxGetClassID(prhs[1])!=mxDOUBLE_CLASS)
    mexErrMsgTxt("I and thresholds must be double arrays.");
  
  const int nelements = (const int) mxGetNumberOfElements(prhs[0]);
  const int nbins     = (const int) mxGetNumberOfElements(prhs[1])-1;
  
  const double *I          = (double*) mxGetPr(prhs[0]);
  const double *thresholds = (double*) mxGetPr(prhs[1]);
  const int do_interp      = (nrhs>=3) ? (const int) mxGetScalar(prhs[2]) : 0;
  const int wrap           = (nrhs>=4) ? (const int) mxGetScalar(prhs[3]) : 0;
  
  plhs[0]      = mxCreateDoubleMatrix(nelements,nbins,mxREAL);
  double *B    = (double*) mxGetPr(plhs[0]);
  plhs[1]      = mxCreateDoubleMatrix(nelements,1,mxREAL);
  double *Bidx = (double*) mxGetPr(plhs[1]);
  
  binning(B, Bidx, I, nelements, thresholds, nbins, do_interp, wrap);
}