%DEMO_CLASSIFICATION - Demonstrate classification for binary images of handwritten digits.

%   Author: Uwe Schmidt, TU Darmstadt (uwe.schmidt@gris.tu-darmstadt.de)
%
%   This file is part of the implementation as described in the CVPR 2012 paper:
%   Uwe Schmidt and Stefan Roth. Learning Rotation-Aware Features: From Invariant Priors to Equivariant Descriptors.
%   Please see the file LICENSE.txt for the license governing this code.

function demo_classification

  %% check for LIBSVM
  assert(exist('svmtrain') == 3 && exist('svmpredict') == 3, ...
         'This demo requires LIBSVM for classification (tested with version 3.1). You can obtain it from: http://www.csie.ntu.edu.tw/~cjlin/libsvm/') %#ok

  %% check for MEX files
  if exist('polarpooling') ~= 3 || exist('binning_fast') ~= 3 %#ok
    fprintf('Compiling MEX files ...\n')
    makemex;
  end

  %% load learned RC-RBM model
  load('models/model_mnist.mat','model'); % model_show(model); pause
  rcrbm_features_    = @(varargin)rcrbm_features(model,varargin{:});
  gradient_features_ = @gradient_features;

  %% choose features to be used
  feature_selection = { rcrbm_features_, gradient_features_ };
  % feature_selection = { rcrbm_features_ };
  % feature_selection = { gradient_features_ };

  %% choose descriptor
  descriptor = @ihof;
  % descriptor = @ehof;

  %% classification pipeline
  fprintf('Loading data ...\n')
  [imdims, data_train,labels_train, data_test,labels_test, data_val,labels_val] = get_data();
  
  fprintf('Computing features on validation set ...\n')
  features_val = final_features(feature_selection,descriptor,imdims,data_val);

  fprintf('Cross validation for SVM Gaussian kernel parameters:\n')
  [c,g] = crossvalidation_svm(features_val,labels_val);
  fprintf('Best parameters: c = %g, g = %.4f\n', c, g)
    
  fprintf('Computing features on train & test set ...\n')
  [features_train,features_test] = final_features(feature_selection,descriptor,imdims,data_train,data_test);
  
  fprintf('SVM training ...\n')
  svm = svmtrain(labels_train, features_train, sprintf('-t 2 -q -c %f -g %f', c, g));
  
  fprintf('SVM prediction on test set ...\n')
  svmpredict(labels_test, features_test, svm);
end


function features = rcrbm_features(rcrbm,data,imdims,descriptor)
  unsigned               = false;
  nrings                 = 2;
  ncells_per_ring        = rcrbm.ntforms;
  neighbor_normalization = false;

  nimages = size(data,2);
  cliques_idx = im2col(reshape(1:prod(imdims),imdims), rcrbm.fdims, 'sliding');
  for i = 1:nimages
    if mod(i,100)==0, fprintf('RC-RBM features:   image %4d/%4d\r',i,nimages), end
    I      = reshape(data(:,i),imdims);
    binned = rcrbm_activation_binning(I(cliques_idx),imdims,rcrbm,unsigned);
    H      = descriptor(binned,nrings,ncells_per_ring,neighbor_normalization);
    if i == 1, features = zeros(nimages,numel(H)); end
    features(i,:) = H(:);
  end
  fprintf('\n')
end

function features = gradient_features(data,imdims,descriptor)
  unsigned               = false;
  nrings                 = 2;
  ncells_per_ring        = 8;
  neighbor_normalization = true;

  nimages = size(data,2);
  for i = 1:nimages
    if mod(i,100)==0, fprintf('Gradient features: image %4d/%4d\r',i,nimages), end
    I      = reshape(data(:,i),imdims);
    binned = gradient_binning(I,ncells_per_ring,unsigned);
    H      = descriptor(binned,nrings,ncells_per_ring,neighbor_normalization);
    if i == 1, features = zeros(nimages,numel(H)); end
    features(i,:) = H(:);
  end
  fprintf('\n')
end

function [features_train,features_test] = final_features(feature_selection,descriptor,imdims,data_train,data_test)
  has_test = exist('data_test','var');
  [features_train,features_test] = deal([]);
  for i = 1:numel(feature_selection)
                 features_train = [features_train,feature_selection{i}(data_train,imdims,descriptor)];     %#ok
    if has_test, features_test  = [features_test, feature_selection{i}(data_test, imdims,descriptor)]; end %#ok 
  end
  % feature normalization (divide each dim by its max value)
  scaling = max(features_train,[],1);
               features_train = bsxfun(@rdivide,features_train,scaling);
  if has_test, features_test  = bsxfun(@rdivide,features_test, scaling); end
end

function [c,g] = crossvalidation_svm(features_val,labels_val)
  cs = 2.^(-0:2);
  gs = 2.^(-4:1);
  [cs,gs] = ndgrid(cs,gs);  
  acc = zeros(1,numel(gs));
  for i = 1:numel(gs)
    fprintf('c = %g, g = %.4f, ', cs(i), gs(i))
    acc(i) = svmtrain(labels_val, features_val, sprintf('-t 2 -q -v 4 -c %f -g %f', cs(i), gs(i)));
  end
  [~,maxidx] = max(acc);
  c = cs(maxidx(1));
  g = gs(maxidx(1));
end