%HOF - Shared implementation of EHOF and IHOF descriptors.
%
%   See also EHOF and IHOF.

%   Author: Uwe Schmidt, TU Darmstadt (uwe.schmidt@gris.tu-darmstadt.de)
%
%   This file is part of the implementation as described in the CVPR 2012 paper:
%   Uwe Schmidt and Stefan Roth. Learning Rotation-Aware Features: From Invariant Priors to Equivariant Descriptors.
%   Please see the file LICENSE.txt for the license governing this code.

function H = hof(binned,nrings,ncells_per_ring,do_normalization,post)

  % recursively call function for each input feature map (and return a big vector)
  if size(binned,4) > 1
    H = arrayfun(@(i) {reshape(hof(binned(:,:,:,i),nrings,ncells_per_ring,do_normalization,post),[],1)}, 1:size(binned,4));
    H = vertcat(H{:});
    return
  end

  do_interp = true;
  nbins = size(binned,3);  
  assert(mod(ncells_per_ring,nbins)==0,'ncells_per_ring should be a multiple of nbins');
  
  cells = polarpooling(binned,nrings,ncells_per_ring,0,do_interp);
  cells_norm = zeros((nrings-1)*ncells_per_ring,nbins+nbins*do_normalization);

  if do_normalization
    nnorms_idx = zeros(2,nbins); for i = 1:2, nnorms_idx(i,:) = i:2:2*nbins; end
    for i = 2:nrings
      for j = 1:ncells_per_ring
        % normalize with predecessor and successor cell on the ring
        if j == 1, pred = ncells_per_ring; else pred = j-1; end
        if j == ncells_per_ring, succ = 1; else succ = j+1; end
        cells_norm((i-1)+(j-1)*(nrings-1), nnorms_idx(1,:)) = norm_fun( cells(i,j,:), cells(i,[pred,j],:) );
        cells_norm((i-1)+(j-1)*(nrings-1), nnorms_idx(2,:)) = norm_fun( cells(i,j,:), cells(i,[succ,j],:) );
      end
    end
  else
    for i = 2:nrings
      for j = 1:ncells_per_ring
        cells_norm((i-1)+(j-1)*(nrings-1),:) = cells(i,j,:);
      end
    end
  end

  H = post(cells,cells_norm);

end

function r = norm_fun(v,b)
  r = v / sqrt(  sum(b(:).^2)   + 1e-4  ); % l2 normalization
end