//POLARPOOLING - 2D value pooling (max or sum) on polar coordinate system (supports bilinear interpolation)
//
//   Compile with MAKEMEX.
//
//   Author: Uwe Schmidt, TU Darmstadt (uwe.schmidt@gris.tu-darmstadt.de)
//
//   This file is part of the implementation as described in the CVPR 2012 paper:
//   Uwe Schmidt and Stefan Roth. Learning Rotation-Aware Features: From Invariant Priors to Equivariant Descriptors.
//   Please see the file LICENSE.txt for the license governing this code.

#include <math.h>
#include "mex.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define pi M_PI

static inline double max(double x, double y) { return (x <= y ? y : x); }
static inline int sign(double x) { return (x < 0 ? -1 : (x > 0 ? 1 : 0)); }


mxArray* pooling(const double *I, const int h, const int w, const int nchannels, const int nrings, const int ncells_per_ring, const int op, const int do_interp) {

  // all cell centers have the same distance in both radius and angle
  const double ring_height = 1.0/(2.0*nrings-1.0);     // height for rings except central one
  const double cc          = ring_height/2.0;          // radius for central ring
  const double cell_width  = (2.0*pi)/ncells_per_ring; // angle steps
  
  double * radii  = (double*) mxMalloc(nrings * sizeof(double));
  for (int i = 0; i < nrings; i++) radii[i] = cc + i*ring_height;

  double * angles  = (double*) mxMalloc(ncells_per_ring * sizeof(double));
  for (int i = 0; i < ncells_per_ring; i++) angles[i] = -pi + (i+1)*cell_width;
    
  const int Hdims[3] = {nrings,ncells_per_ring,nchannels};
  mxArray *mxH = mxCreateNumericArray(3, Hdims, mxDOUBLE_CLASS, mxREAL);
  double *H = (double*) mxGetPr(mxH);
  
  // loop over all pixels
  for (int py = 0; py < h; py++) for (int px = 0; px < w; px++) {
    
    // relative pixel coordinates (0 to 1)
    const double yr = (py+0.5)/h;
    const double xr = (px+0.5)/w;
    // relative to center of image
    const double yy = yr-0.5;
    const double xx = xr-0.5;
    // convert to polar coordinates (angle t, radius r)
    const double t = atan2(yy,xx);
    const double r = sqrt(yy*yy + xx*xx);
    
    // naive classification of ring and section
    int ring = nrings; int section = -1;
    for (int i = 0; i < nrings; i++) {
      if (r <= radii[i]) { ring = i; break; }
    }
    for (int i = 0; i < ncells_per_ring; i++) {
      if (t <= angles[i]) { section = i; break; }
    }
    
    // if pixel is inside the circle (otherwise ignore)
    if (ring < nrings) {
      
      // no interpolation
      if (!do_interp) {
        for (int i = 0; i < nchannels; i++) {
          const double *src = I + i* w* h + px* h + py;
                double *dst = H + i*nrings*ncells_per_ring + (ring!=0)*section*nrings + ring;
          if      (op == 0) *dst =      *dst + *src  ;
          else if (op == 1) *dst = max( *dst,  *src );
        }
        
      // bilinear interpolation in polar coordinates
      } else {
        
        // cell center
        const double r_center = ring == 0 ? 0 : radii[ring] - ring_height/2;
        const double t_center = ring == 0 ? 0 : angles[section] - cell_width/2;      
        // relative position withing cell (wrt. cell center)
        const double crr = (r - r_center) / ring_height;
        const double ctr = (t - t_center) / cell_width; // not defined in circle center
        // distance from cell center
        const double crrd = fabs(crr);
        const double ctrd = ring == 0 ? 0 : fabs(ctr); // not defined in central cell
        // offset to neighboring cells
        const int r_neighbor_offset = sign(crr);
        const int t_neighbor_offset = sign(ctr);
        
        const int ring_neighbor = ring + r_neighbor_offset;
        const int section_neighbor = (ncells_per_ring + section + t_neighbor_offset) % ncells_per_ring;
        
        for (int i = 0; i < nchannels; i++) {
          // contents of image at current pixel and channel
          const double *src = I + i*w*h + px*h + py;
          // base address of i-th channel of H
          double *dst = H + i*nrings*ncells_per_ring;
          
          // special case of central cell
          if (ring == 0) {
            // main cell
            *dst += *src * (1-crrd)*(1-ctrd);
            // ring neighbor only, section neighbor not defined
            if (ring_neighbor < nrings) {
              *(dst + section*nrings + ring_neighbor) += *src * (crrd)*(1-ctrd);
            }
          // all other cells
          } else {
            // main cell
            *(dst + section          * nrings + ring) += *src * (1-crrd)*(1-ctrd);
            // there's always a section neighbor (radius stays the same)
            *(dst + section_neighbor * nrings + ring) += *src * (1-crrd)*(ctrd);
            // special case of ring neighbor being central cell
            if (ring_neighbor == 0) {
              *dst += *src * (crrd)*(1-ctrd);
            // otherwise ring and section neighbor
            } else if (ring_neighbor < nrings) {
              *(dst + section          * nrings + ring_neighbor) += *src * (crrd)*(1-ctrd);
              *(dst + section_neighbor * nrings + ring_neighbor) += *src * (crrd)*(ctrd);
            }
          }
        }
      }
    }
  }
  
  mxFree(radii);
  mxFree(angles);

  return mxH;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  
  if( nrhs<3 || nrhs>5 ) mexErrMsgTxt("3-5 inputs expected.");
  if( nlhs>1 ) mexErrMsgTxt("One output expected.");
  
  const int nDims = mxGetNumberOfDimensions(prhs[0]);
  const int *dims = mxGetDimensions(prhs[0]);
  const int nCh = (nDims==2) ? 1 : dims[2];
  if( (nDims!=2 && nDims!=3) || mxGetClassID(prhs[0])!=mxDOUBLE_CLASS)
    mexErrMsgTxt("I must be a double array with 2 or 3 dimensions.");
  
  const double *I           = (double*) mxGetPr(prhs[0]);
  const int nrings          = (int) mxGetScalar(prhs[1]);
  const int ncells_per_ring = (int) mxGetScalar(prhs[2]);
  const int op              = (nrhs>=4) ? (int) mxGetScalar(prhs[3]) : 0;
  const int do_interp       = (nrhs>=5) ? (int) mxGetScalar(prhs[4]) : 1;
  
  if (op != 0 && op != 1)
    mexErrMsgTxt("op must either be 0 (sum) or 1 (max).");
    
  if (nrings == 1 && ncells_per_ring != 1) {
    mexErrMsgTxt("Only single cell for single ring at the moment.\n");
  }
  
  if (do_interp && ncells_per_ring == 1) {
    mexErrMsgTxt("Need more than 1 cell per ring with bilinear interpolation.");
  }
  
  if (do_interp && op == 1) {
    mexPrintf("Interpolation not used for max-pooling.\n");
  }
  
  plhs[0] = pooling(I, dims[0], dims[1], nCh, nrings, ncells_per_ring, op, do_interp);  
}